package dev.peerat.parser.java;

import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.jupiter.api.Assertions.assertEquals;

import java.util.List;
import java.util.function.Function;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.TestInstance.Lifecycle;

import dev.peerat.parser.Parser;
import dev.peerat.parser.java.tree.JavaTreeType;
import dev.peerat.parser.java.value.BiValue;
import dev.peerat.parser.java.value.MethodCallValue;
import dev.peerat.parser.java.value.ModifierValue;
import dev.peerat.parser.java.value.StaticValue;
import dev.peerat.parser.java.value.TriValue;
import dev.peerat.parser.java.value.Value;
import dev.peerat.parser.java.value.Value.ValueContainer;

@TestInstance(Lifecycle.PER_CLASS)
public class ValueTypesTests {
	
	private Parser<JavaElement> parser;
	
	@BeforeAll
	public void init(){
		parser = new JavaParser(JavaTreeType.VALUE);
	}
	
	private Value parse(String value) throws Exception{
		Container container = new Container();
		parser.parse(value, container);
		return container.getValue();
	}
	
	private static class Container extends JavaElement implements ValueContainer{
		
		private Value value;
		
		@Override
		public void addValue(Value value) {
			this.value = value;
		}
		
		public Value getValue(){
			return this.value;
		}

		@Override
		public void build(Builder builder) throws Exception {}

		@Override
		public <E extends JavaElement> E find(Function<JavaElement, Boolean> finder){return null;}

		@Override
		public <E extends JavaElement> void findAll(Function<JavaElement, Boolean> finder, List<E> list){}
		
	}
	
	private <T> T assertInstance(Object obj, java.lang.Class<? extends T> type){
		assertTrue(type.isAssignableFrom(obj.getClass()));
		return (T) obj;
	}

	@Test
	public void onStaticValue() throws Exception{
		
		//ALL FORMATED VALUES LIKE DOUBLE, FLOAT, BINARY, CHARS(?),.....
		
		StaticValue v;
		
		v = assertInstance(parse("8"), StaticValue.class);
		assertEquals("8", v.getToken().getValue());		
		v = assertInstance(parse("value"), StaticValue.class);
		assertEquals("value", v.getToken().getValue());	
		v = assertInstance(parse("\"value\""), StaticValue.class);
		assertEquals("\"value\"", v.getToken().getValue());
		v = assertInstance(parse("'a'"), StaticValue.class);
		assertEquals("'a'", v.getToken().getValue());
	}
	
	@Test
	public void onBiValue() throws Exception{
		
		BiValue v;
		
		v = assertInstance(parse("8+8"), BiValue.class);
		assertEquals("8", assertInstance(v.left(), StaticValue.class).getToken().getValue());
		assertEquals("+", v.getAction());
		assertEquals("8", assertInstance(v.right(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("8+offset"), BiValue.class);
		assertEquals("8", assertInstance(v.left(), StaticValue.class).getToken().getValue());
		assertEquals("+", v.getAction());
		assertEquals("offset", assertInstance(v.right(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("8+offset > 8"), BiValue.class);
		BiValue right = assertInstance(v.right(), BiValue.class);
		assertEquals("8", assertInstance(v.left(), StaticValue.class).getToken().getValue());
		assertEquals("+", v.getAction());
		assertEquals("offset", assertInstance(right.left(), StaticValue.class).getToken().getValue());
		assertEquals(">", right.getAction());
		assertEquals("8", assertInstance(right.right(), StaticValue.class).getToken().getValue());
	}
	
	@Test
	public void onTriValue() throws Exception{
		
		TriValue v;
		
		v = assertInstance(parse("bool ? true : false"), TriValue.class);
		assertEquals("bool", assertInstance(v.getChecker(), StaticValue.class).getToken().getValue());
		assertEquals("true", assertInstance(v.success(), StaticValue.class).getToken().getValue());
		assertEquals("false", assertInstance(v.fail(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("bool ? 8+8 : false"), TriValue.class);
		assertEquals("bool", assertInstance(v.getChecker(), StaticValue.class).getToken().getValue());
		BiValue bv = assertInstance(v.success(), BiValue.class);
		assertEquals("8", assertInstance(bv.left(), StaticValue.class).getToken().getValue());
		assertEquals("+", bv.getAction());
		assertEquals("8", assertInstance(bv.right(), StaticValue.class).getToken().getValue());
		assertEquals("false", assertInstance(v.fail(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("bool ? rebool ? true : bool : false"), TriValue.class);
		assertEquals("bool", assertInstance(v.getChecker(), StaticValue.class).getToken().getValue());
		assertEquals("false", assertInstance(v.fail(), StaticValue.class).getToken().getValue());
		TriValue success = assertInstance(v.success(), TriValue.class);
		assertEquals("rebool", assertInstance(success.getChecker(), StaticValue.class).getToken().getValue());
		assertEquals("true", assertInstance(success.success(), StaticValue.class).getToken().getValue());
		assertEquals("bool", assertInstance(success.fail(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("(rebool ? true : bool) ? true : false"), TriValue.class);
		TriValue checker = assertInstance(v.getChecker(), TriValue.class);
		assertEquals("rebool", assertInstance(checker.getChecker(), StaticValue.class).getToken().getValue());
		assertEquals("true", assertInstance(checker.success(), StaticValue.class).getToken().getValue());
		assertEquals("bool", assertInstance(checker.fail(), StaticValue.class).getToken().getValue());
		assertEquals("true", assertInstance(v.success(), StaticValue.class).getToken().getValue());
		assertEquals("false", assertInstance(v.fail(), StaticValue.class).getToken().getValue());
	}
	
	@Test
	public void onModifierValue() throws Exception{
		
		ModifierValue v;
		v = assertInstance(parse("!a"), ModifierValue.class);
		assertEquals("!", v.getModifier().getValue());
		assertEquals("a", assertInstance(v.getValue(), StaticValue.class).getToken().getValue());
		
		v = assertInstance(parse("!(a || b)"), ModifierValue.class);
		assertEquals("!", v.getModifier().getValue());
		BiValue bv = assertInstance(v.getValue(), BiValue.class);
		assertEquals("a", assertInstance(bv.left(), StaticValue.class).getToken().getValue());
		assertEquals("||", bv.getAction());
		assertEquals("b", assertInstance(bv.right(), StaticValue.class).getToken().getValue());
	}
	
	@Test
	public void onMethodCallValue() throws Exception{
		MethodCallValue v;
		v = assertInstance(parse("test()"), MethodCallValue.class);
		assertNull(v.base());
		assertEquals("test", v.getToken().getValue());
		assertNull(v.getGeneric());
		assertNull(v.getParameters());
		
		v = assertInstance(parse("JavaParserTest.test(8+8).equals(16)"), MethodCallValue.class);
		assertEquals("equals", v.getToken().getValue());
		assertEquals(1, v.getParameters().size());
		assertEquals("16", assertInstance(v.getParameters().get(0), StaticValue.class).getToken().getValue());
		v = assertInstance(v.base(), MethodCallValue.class);
		assertEquals(1, v.getParameters().size());
		BiValue bv = assertInstance(v.getParameters().get(0), BiValue.class);
		assertEquals("8", assertInstance(bv.left(), StaticValue.class).getToken().getValue());
		assertEquals("+", bv.getAction());
		assertEquals("8", assertInstance(bv.right(), StaticValue.class).getToken().getValue());
		assertEquals("test", v.getToken().getValue());
		assertEquals("JavaParserTest", assertInstance(v.base(), StaticValue.class).getToken().getClass());
		
		v = assertInstance(parse("main.test.here.localvar.run(6)[0].validate(7)"), MethodCallValue.class);
	}
}
